
import inox._
import inox.trees.{forall => _, _}
import inox.trees.dsl._

import welder._

object FoldContent {

  /* We start by defining the relevant functions and datatypes as in the fold left/right case. */

  val foldID = FreshIdentifier("fold")
  val contentID = FreshIdentifier("content")

  val list = FreshIdentifier("List")
  val cons = FreshIdentifier("Cons")
  val nil = FreshIdentifier("Nil")

  val head = FreshIdentifier("head")
  val tail = FreshIdentifier("tail")

  val listADT = mkSort(list)("A")(Seq(cons, nil))
  val consADT = mkConstructor(cons)("A")(Some(list)) { case Seq(aT) =>
    Seq(ValDef(head, aT), ValDef(tail, T(list)(aT)))
  }
  val nilADT = mkConstructor(nil)("A")(Some(list))(_ => Nil)

  /* A definition of fold left */
  val foldFunction = mkFunDef(foldID)("A") { case Seq(aT) => (
    Seq("f" :: ((aT, aT) =>: aT), "z" :: aT, "l" :: T(list)(aT)), aT, { case Seq(f, z, l) =>
      if_ (l.isInstOf(T(cons)(aT))) {
        E(foldID)(aT)(f, f(z, l.asInstOf(T(cons)(aT)).getField(head)), l.asInstOf(T(cons)(aT)).getField(tail))
        } else_ {
          z
        }
    })
  }

  /* The list content function, as defined in the previous exercises in Stainless */
  val contentFunction = mkFunDef(contentID)("A") { case Seq(aT) => (
    Seq("l" :: T(list)(aT)), SetType(aT), { case Seq(l) =>
      if_ (l.isInstOf(T(cons)(aT))) {
        SetAdd(E(contentID)(aT)(l.asInstOf(T(cons)(aT)).getField(tail)), l.asInstOf(T(cons)(aT)).getField(head))
        } else_ {
          FiniteSet(Seq.empty, aT)
        }
    })
  }

  val symbols = NoSymbols
    .withADTs(Seq(listADT, consADT, nilADT))
    .withFunctions(Seq(foldFunction, contentFunction))

  val program = InoxProgram(Context.empty, symbols)

  val theory = theoryOf(program)
  import theory._

  val A = TypeParameter.fresh("A")

  def fold(f: Expr, z: Expr, l: Expr) = E(foldID)(A)(f, z, l)
  def content(e: Expr) = E(contentID)(A)(e)

  val ListA = T(list)(A)
  val ConsA = T(cons)(A)
  val NilA = T(nil)(A)


  /* TODO: state and prove the following theorem:
   *
   *    ∀f: (A,A) => A,z: A, l1: List[A], l2: List[A]. (
   *      isAssociative(f) &&
   *      isCommutative(f) &&
   *      isIdempotent(f) &&
   *      content(l1) == content(l2)
   *    ) ==> (fold(f, z, l1) == fold(f, z, l2))
   */
  lazy val theorem: Theorem = ???
}
